import subprocess
from . import base
from . import exceptions
import tuned.logs
from tuned.utils.commands import commands

log = tuned.logs.get()

ACTION_TIMEOUT = 180

class ApplicationPlugin(base.Plugin):
	"""
	`application`:

	Dynamically executes the optimization action according to the application 
	running situation.
	"""
	last_apps_running = set()

	def __init__(self, *args, **kwargs):
		super(ApplicationPlugin, self).__init__(*args, **kwargs)
		self._has_dynamic_options = True

	def _instance_init(self, instance):
		instance._has_static_tuning = False
		instance._has_dynamic_tuning = True
		# save all options
		instance.option_dict = {}
		for option, value in list(instance.options.items()):
			instance.option_dict[option] = value
			log.debug(f"{option}: {value}")
		instance._load_monitor = self._monitors_repository.create("process", None)

	def _instance_cleanup(self, instance):
		if instance._load_monitor is not None:
			self._monitors_repository.delete(instance._load_monitor)
			instance._load_monitor = None
		instance.option_dict.clear()

	def _instance_update_dynamic(self, instance, device):
		if "name" not in instance.option_dict:
			return
		applications_monitored = set([app.strip() for app in instance.option_dict["name"].split(',')])
		if not applications_monitored:
			return
		apps_running = applications_monitored.intersection(instance._load_monitor.get_load()["application"])
		log.debug("running: " + str(apps_running))
		new_apps = apps_running - self.last_apps_running
		gone_apps = self.last_apps_running - apps_running
		if len(new_apps) > 0:
			log.info("new apps: " + str(new_apps))
		if len(gone_apps) > 0:
			log.info("gone apps: " + str(gone_apps))
		for app in gone_apps:
			self._execute_action(instance, app, rollback=True)
		for app in new_apps:
			self._execute_action(instance, app, rollback=False)
		self.last_apps_running = apps_running

	def _instance_unapply_dynamic(self, instance, device):
		# restore previous value
		pass

	def _execute_action(self, instance, app, rollback=False):
		# find action
		if rollback:
			option = f"{app}_rollback_action"
		else:
			option = f"{app}_action"
		if option not in instance.option_dict:
			return
		action = str(instance.option_dict[option])
		# remove wrapping " or ' in action
		if len(action) >= 2 and (action[0] == '"' or action[0] == "'") and action[0] == action[-1]:
			action = action[1:-1]
		# execute action for app
		if len(action):
			log.info(f"{option}: {action}")
			try:
				p = subprocess.Popen(action, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
				retval = p.wait(ACTION_TIMEOUT)
				log.info(f"{option}: return {retval}")
			except Exception as e:
				log.info(f"{option}: {str(e)}")
